/*
 * stosoo_template.c
 *
 * Version 0.1 (11/11/2013)
 *
 * File created on Nov. 11th, 2013.
 *
 * Copyright 2013, 2014 Philippe Preux
 * Philippe.Preux@univ-lille3.fr
 *
 * This is GPL software.
 *
 * Implementation of the StoSOO algorithm for the optimization of
 * noisy functions.
 *
 * Based on Valko, Carpentier, Munos, ICML 2013 paper.
 *
 * In some cases, we may use slightly less calls to the objective function
 * than the prescribed number. In any case, we do not use more than the
 * authorized budget.
 *
 * This file is part of yaStoSOO.

    yaStoSOO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    yaStoSOO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with yaStoSOO.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

/* for getopt() arguments */
#include <ctype.h>
#include <getopt.h>

#include <errno.h>

#include "print.help.h"
#include "stosoo_optim.h"
#include "Malloc.h"

#define MY_NAME program_invocation_short_name
#define VERSION "0.3"
#define defaultOutputFilename "output"

#define square(x) ((x)*(x))

int verbose_mode = 0;

static unsigned int seed = 1; /* default seed for random() */
static size_t d = 2;

#ifdef ENABLE_NLOPT
static int post_processing = 0;
#endif

extern void set_noise_std_dev (const double);

static void init (void)
{
  srand (seed);
}

static void analParam (int argc, char *argv[])
{
  static struct option long_options[] =
    {
      {"dim", 1, 0, 1},
      {"ns", 1, 0, 2},
      {"nv", 1, 0, 3},
      {"N_noeuds", 1, 0, 6},
#ifdef ENABLE_NLOPT
      {"post-process", 1, 0, 9},
      {"stosoo-part", 1, 0, 10},
#endif
      {"seed", 1, 0, 4},
      {"split", 1, 0, 5},
      {"help", 0, 0, 'h'},
      {"verbose", 0, 0, 'V'},
      {"version", 0, 0, 'v'},
      {0, 0, 0, 0}
    };
  FILE *flog;
  int i;

  flog = fopen (".stosoo.log", "w");
  for (i = 0; i < argc; i ++)
    fprintf (flog, "%s ", argv [i]);
  fprintf (flog, "\n");
  fclose (flog);

  while (1) {
    int option_index = 0;
    char c = getopt_long (argc, argv, "K:k:N:d:D:hVv",
		     long_options, &option_index);
    if (c == -1) break;
    switch (c) {
    case 1:
      d = (size_t) atol (optarg);
      break;
    case 2: {
      double v = atof (optarg);
      if (v < 0) {
	fprintf (stderr, "%s: a standard deviation is positive!\n", MY_NAME);
      } else set_noise_std_dev (v);
      break;
    }
    case 3: {
      double v = atof (optarg);
      if (v < 0) {
	fprintf (stderr, "%s: a standard deviation is positive!\n", MY_NAME);
      } else set_noise_std_dev (sqrt (v));
      break;
    }
    case 4:
      seed = (unsigned int) atoi (optarg);
      break;
    case 5:
      if (! strcmp (optarg, "increasing"))
	yaStoSOO_set_split_policy (1);
      else if (! strcmp (optarg, "decreasing"))
	yaStoSOO_set_split_policy (-1);
      else if (! strcmp (optarg, "random"))
	yaStoSOO_set_split_policy (0);
      break;
    case 6:
      yaStoSOO_set_N_noeuds ((size_t) atol (optarg));
      break;
#ifdef ENABLE_NLOPT
    case 9:
      if (! strcmp (optarg, "bobyqa"))
	yaStoSOO_set_post_process (1);
      else if (! strcmp (optarg, "sbplx"))
	yaStoSOO_set_post_process (2);
      post_processing = 1;
      break;
    case 10: {
      double v = atof (optarg);
      if ((v >= 0) && (v <= 1))
	yaStoSOO_set_percentage (atof (optarg));
      else {
	fprintf (stderr, "%s: incorrect value (%s)\n", MY_NAME, optarg);
	exit (EXIT_FAILURE);
      }
    }
#endif
    case 'D':
      yaStoSOO_set_max_depth ((size_t) (atol (optarg)));
      break;
    case 'd':
      yaStoSOO_set_delta (atof (optarg));
      break;
    case 'K': {
      int v = atol (optarg);
      if (v < 0) {
	fprintf (stderr, 
		 "%s: split factor should be positive (%s provided)\n", 
		 MY_NAME, optarg);
	exit (EXIT_FAILURE);
      } else if (v % 2 != 1) {
	fprintf (stderr, 
		 "%s: split factor should be odd (%s provided)\n", 
		 MY_NAME, optarg);
	exit (EXIT_FAILURE);
      }
      yaStoSOO_set_split_factor ((size_t) v);
      break;
    }
    case 'k':
      yaStoSOO_set_max_number_of_evaluations_per_node ((size_t) (atol (optarg)));
      break;
    case 'N':
      yaStoSOO_set_number_of_obj_fct_evaluations ((size_t) (atol (optarg)));
      break;
    case 'h':
      print_help (long_options);
      exit (EXIT_SUCCESS);
      break;
    case 'V':
      verbose_mode = 1;
      break;
    case 'v':
      fprintf (stderr, "%s: version %s\n", MY_NAME, VERSION);
      exit (EXIT_SUCCESS);
    }
  }
}

extern void print_optimum (void);
extern double f_maximum (void);

int main (int argc, char *argv [])
{
  analParam (argc, argv);
  init ();

  /* The domain is a compact subset of R^d, d being the value of the 
     static variable. */
  double *or, *dim; /* origins and dimensions of the domain */
  or = Malloc (sizeof (double) * d);
  dim = Malloc (sizeof (double) * d);
  for (size_t i = 0; i < d; i ++) { or [i] = 0.; dim [i] = 1.; }
  yaStoSOO_set_domain (d, or, dim);

  yaStoSOO_optim_init ();
  yaStoSOO_optim ();

  t_les_x_star_val *les_res = yaStoSOO_les_x_star_val ();
  double f_star = f_maximum ();

  for (size_t i = 0; i < yaStoSOO_get_N_values (); i ++) {
    printf ("%g", les_res [i]. x_star_preproc [0]);
    for (size_t j = 1; j < d; j ++)
      printf (" %g", les_res [i]. x_star_preproc [j]);
    printf (": %g", les_res [i]. x_star_val_preproc);
    if (f_star == f_star) /* cheap trick to test for not being a NaN */
      printf (" %g\n", fabs (f_star - les_res [i]. x_star_val_preproc));
    else
      printf (" NA\n");
  }
  return EXIT_SUCCESS;
}
