/*
 * stosoo_optim.c
 *
 * Copyright 2013, 2014 Philippe Preux
 * Philippe.Preux@univ-lille3.fr
 *
 * This is GPL software.
 *
 * Implementation of the StoSOO algorithm for the optimization of
 * noisy functions.
 *
 * Based on Valko, Carpentier, Munos, ICML 2013 paper.
 *
 * It some cases, we may use slightly less calls to the objective function
 * than the prescribed number. In any case, we do not use more than the
 * authorized budget.
 *
 * To do:
   - manage hmin
   - write a specific code for d = 1
   - remove from the main loop the processing of the root node
 *
 * v 0.3:
 * Apr 8th, 2014.
 * Considerable speed-up (50 x) obtained by using the leaves data structure.
 * Apr 9th, 2014.
 * Fix bug in the management of the direction of splits.
 * Now, by default, splits cells as expected according to SOO definition.
 *
 *
 * v 0.2:
 * Apr 8th, 2014.
 * Correction of a bug in function add_note_to_leaves()
 * Add
       add_node_to_leaves (0);
     at the end of stosoo_optim_init() instead of
       nb_leaves [0] = 1;
 * Modification in split_node(): 
       nodes [node_id]. expanded = 1:
     instead of
       nodes [node_id]. expanded = nb_nodes;
 * Make split_node () a void function.
 *
 * This file is part of yaStoSOO.

    yaStoSOO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    yaStoSOO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with yaStoSOO.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef ENABLE_NLOPT
#include "nlopt.h"
#endif

#include "stosoo_optim.h"
#include "Malloc.h"

extern const size_t nb_calls_to_obj_fct;

static size_t N = 0; /* Number of objective function calls (aka maxeval) */
static size_t K = 3;

/* The 3 next variables are generally set with default (smart) values */
static size_t k = 0;
static size_t hmin = 0;
static size_t hmax = 0;
static double delta = .0;

static size_t current_hmax = 0;
static int split_policy = 1;
/*
static double **dim_sizes = NULL;
static size_t dim_sizes_nb_elt = 0;
*/
static int soo_flag = 0;

void yaStoSOO_set_hmax (const size_t value)
{
  hmax = value;
}

void yaStoSOO_set_split_policy (const int p)
{
  split_policy = p;
}

/* Regarding the domain of the objective function */
/*static*/ size_t d = 0; /* Dimension of the domain */
static double *domain_origins = NULL;
static double *domain_dimensions = NULL;

/* Data structures to represent the tree. */
static P_noeud nodes = NULL;
static size_t nb_nodes = 0;

size_t yaStoSOO_nb_nodes (void)
{
  return nb_nodes;
}

/* Data structure for the leaves */
#define N_new_leaves 16384 /*10000*/
static size_t **leaves = NULL;
static size_t *nb_leaves = 0;
static size_t *size_leaves = 0;

#ifdef PROGRESSION
static size_t **progression = NULL;
#endif

/* The optimum and its value. */
static double *x_star = NULL;
static double x_star_val = -HUGE_VAL;

extern int verbose_mode;

static void add_node_to_leaves (size_t node_id)
{
  size_t depth = nodes [node_id]. depth;
  if (nb_leaves [depth] == size_leaves [depth]) {
    size_leaves [depth] += N_new_leaves;
    leaves [depth] = 
      Realloc (leaves [depth], sizeof (size_t) * size_leaves [depth]);
    if (verbose_mode) fprintf (stderr, "# Realloc %ld\n", size_leaves [depth]);
  }
  /*  printf ("leaves [%d] [nb_leaves [%ld] == %ld] <- %ld\n", 
      depth, depth, nb_leaves [depth], node_id);*/
  leaves [depth] [nb_leaves [depth]] = node_id; /* !!size_leaves -> nb_leaves*/
  nb_leaves [depth] ++;
}
/*
static void remove_node_from_leaves (size_t node_id)
{
}
*/

#ifdef ENABLE_NLOPT
static nlopt_algorithm post_process_algo = NLOPT_LN_BOBYQA;
void yaStoSOO_set_post_process (const int algo)
{
  if (algo == 1) post_process_algo = NLOPT_LN_BOBYQA;
  else if (algo == 2) post_process_algo = NLOPT_LN_SBPLX;
}
#endif

static double stosoo_percentage = 1.0;

void yaStoSOO_set_percentage (const double v)
{
  stosoo_percentage = v;
}

static size_t N_noeuds = 100000;

void yaStoSOO_set_N_noeuds (const size_t value)
{
  N_noeuds = value;
}

void yaStoSOO_print_tree (void)
{
  char *padding_string = Malloc (sizeof (char) * hmax + 1);
  for (size_t depth = 0; depth < hmax; depth ++) {
    padding_string [depth] = '\0';
    for (size_t node = 0; node < nb_nodes; node ++)
      if (nodes [node]. depth == depth) {
	printf ("%s%ld %ld %ld", padding_string, node, 
		nodes [node]. parent, nodes [node]. expanded);
	for (size_t i = 0; i < d; i ++) printf (" %f", nodes [node]. x [i]);
	for (size_t i = 0; i < d; i ++) printf (" %f", nodes [node]. dim [i]);
	printf ("\n");
      }
    padding_string [depth] = ' ';
  }
}

extern double f (DataPoint);
extern double f_init (const size_t, const size_t);

#define min(x,y) ((x)<(y)?(x):(y))

static size_t *nodes_to_split_list;
static size_t nb_nodes_in_the_split_list = 0;

void yaStoSOO_set_soo_mode (void)
{
  soo_flag = 1;
  k = 1;
}

void yaStoSOO_optim_init (void)
{
  size_t facteur = 2; /* faudra tuner ce 2 en fonction de k, K, hmax, ...*/
  if (N == 0) N = 100;
  double log_N = log ((double) N); /* cache */
  double log_N_3 = log_N * log_N * log_N;
  if (N_noeuds < facteur * N) yaStoSOO_set_N_noeuds (facteur * N);
  nodes = Malloc (sizeof (Noeud) * N_noeuds);
  nodes_to_split_list = Malloc (sizeof (size_t) * N_noeuds);
  if (!k) {
    k = (size_t) ((double) N / log_N_3);
    if (verbose_mode) fprintf (stderr, "# max # of eval per node = %ld\n", k);
  }
  if (!hmax) {
    hmax = (size_t) sqrt (log_N_3);
    if (soo_flag) hmax *= 10;
  }
  if (verbose_mode) fprintf (stderr, "# hmax = %ld\n", hmax);
  if (delta == .0) delta = 1. / sqrt ((double) N);
  if (verbose_mode) fprintf (stderr, "# delta = %g\n", delta);

  /*
  dim_sizes_nb_elt = ceil (hmax / K);
  dim_sizes = Malloc (sizeof (double) * d);
  for (size_t i = 0; i < d; i ++) {
    dim_sizes [i] = Malloc (sizeof (double) * dim_sizes_nb_elt);
    dim_sizes [i] [0] = domains_dimensions [i];
  }
  for (size_t i = 0; i < d; i ++)
    for (size_t j = 1; j < dim_sizes_nb_elt; j ++)
      dim_sizes [i] [j] = dim_sizes [i] [j - 1] / K;
  */
  /* Create leaves */
  leaves = Malloc (sizeof (size_t *) * hmax);
  nb_leaves = Malloc (sizeof (size_t) * hmax);
  size_leaves = Malloc (sizeof (size_t) * hmax);
#define essaiPP
  /*   50D: c ( 4.56177, 4.56596, 4.47697, 4.56731, 4.58314)
      100D: c (21.6102, 21.1565, 21.168, 21.611, 21.6431)

      essaiPP pas défini :
       50D: c (4.65447, 4.57235, 4.58211, 4.58114, 4.62748)
      100D: c (21.6722, 21.6235, 21.6556, 21.6767, 21.7012)
   */
#ifdef essaiPP
  size_t size = N_new_leaves;
#else
  size_t size = 1;
#endif
  for (size_t depth = 0; depth < hmax; depth ++) {
    leaves [depth] = Malloc (sizeof (size_t) * size);
    nb_leaves [depth] = 0;
    size_leaves [depth] = size;
#ifndef essaiPP
    size *= K;
    if (size > N_new_leaves) size = N_new_leaves;
#endif
  }

#ifdef PROGRESSION
  progression = Malloc (sizeof (size_t *) * hmax);
  for (size_t i = 0; i < hmax; i ++)
    progression [i] = Malloc (sizeof (size_t) * 200);
#endif

  for (size_t i = 0; i < N_noeuds; i ++) {
    nodes [i]. expanded = nodes [i]. parent = 
      nodes [i]. nb_eval =  nodes [i]. sum_eval = 0;
    nodes [i]. x = Malloc (sizeof (double) * d);
    nodes [i]. dim = Malloc (sizeof (double) * d);
    nodes [i]. center = Malloc (sizeof (double) * d);
    nodes [i]. split_direction = 0; /* useless init, just to be clean. */
    nodes [i]. dim_index = Malloc (sizeof (size_t) * d);
    for (size_t j = 0; j < d; j ++) nodes [i]. dim_index [j] = 0;
  }

  /* Initialize the objective function. */
  f_init (N, d);

  /* Creation of the root of the tree. */
  nodes [0]. depth = 0;
  (void) memcpy (nodes [0]. x, domain_origins, sizeof (double) * d);
  (void) memcpy (nodes [0]. dim, domain_dimensions, sizeof (double) * d);
  for (size_t j = 0; j < d; j ++)
    nodes [0]. center [j] = nodes [0]. x [j] + nodes [0]. dim [j] / 2.;
  nodes [0]. sum_eval = f (nodes [0]. center);
  nodes [0]. nb_eval = 1;
  nodes [0]. b = nodes [0]. sum_eval + 
    (soo_flag ? .0 : sqrt (log (N * k / delta) / 2.0));
  nb_nodes = 1;

  /* Mettre la racine dans leaves */
  /* !!removed:   nb_leaves [0] = 1;*/
  add_node_to_leaves ((size_t) 0); /* !!added */
}

static void split_node (const size_t node_id, const size_t split_direction)
{
  /* We check that the node has no been yet expanded */
  if (nodes [node_id]. expanded) return;
  /* We check we have enough space allocated */
  if (nb_nodes + K >= N_noeuds) {
    fprintf (stderr, "Memory exhausted\n");
    exit (EXIT_FAILURE);
  }
  /* K new nodes are created and evaluated once. */
  if (current_hmax < nodes [node_id]. depth + 1)
    current_hmax = nodes [node_id]. depth + 1;

  nodes [node_id]. expanded = 1 /*??? nb_nodes*/;
  double dim_split_node = nodes [node_id]. dim [split_direction] / K;

  for (size_t node = 0; node < K; node ++, nb_nodes ++) {
    nodes [nb_nodes]. depth = nodes [node_id]. depth + 1;
    nodes [nb_nodes]. parent = node_id;
    (void) memcpy (nodes [nb_nodes]. x, nodes [node_id]. x, 
		   sizeof (double) * d);
    nodes [nb_nodes]. x [split_direction] += node * dim_split_node;
    (void) memcpy (nodes [nb_nodes]. dim, nodes [node_id]. dim,
		   sizeof (double) * d);
    nodes [nb_nodes]. dim [split_direction] = dim_split_node;
    (void) memcpy (nodes [nb_nodes]. center, nodes [node_id]. center,
		   sizeof (double) * d);
    nodes [nb_nodes]. center [split_direction] =
      nodes [nb_nodes]. x [split_direction] + dim_split_node / 2.;
    nodes [nb_nodes]. split_direction = split_direction;

    if (node == (K >> 1)) {
      nodes [nb_nodes]. b = nodes [node_id]. b;
      nodes [nb_nodes]. sum_eval = nodes [node_id]. sum_eval;
      nodes [nb_nodes]. nb_eval = nodes [node_id]. nb_eval;
      (void) memcpy (nodes [nb_nodes]. center, nodes [node_id]. center, 
		     sizeof (double) * d);
    } else {
      nodes [nb_nodes]. nb_eval = 1;
      nodes [nb_nodes]. sum_eval = f (nodes [nb_nodes]. center);
      nodes [nb_nodes]. b = nodes [nb_nodes]. sum_eval + 
	(soo_flag ? .0 : sqrt (log (N * k / delta) / 2.0));
    }
    add_node_to_leaves (nb_nodes);
  }
  /*  remove_node_from_leaves (node_id);*/
}

void yaStoSOO_set_split_factor (const size_t value)
{
  K = value;
}

void yaStoSOO_set_max_number_of_evaluations_per_node (const size_t value)
{
  k = value;
}

void yaStoSOO_set_number_of_obj_fct_evaluations (const size_t value)
{
  N = value;
}

void yaStoSOO_set_max_depth (const size_t value)
{
  hmax = value;
}

void yaStoSOO_set_delta (const double value)
{
  delta = value;
}

#ifdef ENABLE_NLOPT
static double *nlopt_ub, *nlopt_lb;
#endif

void yaStoSOO_set_domain (const size_t ddim, const double *origins, 
			  const double *dimensions)
{
  d = ddim;
  domain_origins = Malloc (sizeof (double) * d);
  domain_dimensions = Malloc (sizeof (double) * d);
  (void) memcpy (domain_origins, origins, sizeof (double) * d);
  (void) memcpy (domain_dimensions, dimensions, sizeof (double) * d);
#ifdef ENABLE_NLOPT
  nlopt_lb = Malloc (sizeof (double) * d);
  nlopt_ub = Malloc (sizeof (double) * d);
  (void) memcpy (nlopt_lb, origins, sizeof (double) * d);
  for (size_t i = 0; i < d; i ++)
    nlopt_ub [i] = nlopt_lb [i] + dimensions [i];
#endif
}

static void add_to_split_list (size_t node_id)
{
  nodes_to_split_list [nb_nodes_in_the_split_list ++] = node_id;
}

static void split_nodes (void)
{
  static double rnd_scale = 1. / ((double) RAND_MAX);
  static size_t next_split_direction = 0;

  for (size_t i = 0; i < nb_nodes_in_the_split_list; i ++) {
    size_t id = nodes_to_split_list [i];
    switch (split_policy) {
    case SPLIT_ORIGINAL:
      next_split_direction = nodes [id]. split_direction;
      if (next_split_direction < d - 1) next_split_direction ++;
      else next_split_direction = 0;
      break;
    case SPLIT_BY_INCREASING:
      if (next_split_direction < d - 1) next_split_direction ++;
      else next_split_direction = 0;
      break;
    case SPLIT_BY_DECREASING:
      if (next_split_direction > 0) next_split_direction --;
      else next_split_direction = d - 1;
      break;
    case SPLIT_AT_RANDOM:
      next_split_direction = floor (rand () * rnd_scale * d);
      break;
    }
    split_node (id, next_split_direction);
  }
  nb_nodes_in_the_split_list = 0;
}
#define is_a_leaf(node) (!nodes [node]. expanded)

static t_les_x_star_val *les_x_star_val;
t_les_x_star_val *yaStoSOO_les_x_star_val (void)
{
  return les_x_star_val;
}

#ifdef ENABLE_NLOPT
static double myfunc (__attribute__((unused)) unsigned int n, const double *x, 
		      __attribute__((unused)) double *grad,
		      __attribute__((unused)) void *f_data)
{
  return f (x);
}
#endif

extern void cec14_test_func (double *, double *, int, int, int);

#ifdef ENABLE_NLOPT
static double post_process (double *x_star, const size_t maxeval)
{
  /*  nlopt_result code;*/
  nlopt_opt opt = nlopt_create (post_process_algo, (unsigned) d);
  nlopt_set_lower_bounds (opt, nlopt_lb);
  nlopt_set_upper_bounds (opt, nlopt_ub);
  nlopt_set_max_objective (opt, myfunc, NULL); /* !!! max for cec214 compet. */
  (void) nlopt_set_maxeval (opt, (int) maxeval);
  (void) nlopt_optimize (opt, x_star, &x_star_val);
  return (x_star_val);
}
#endif

typedef struct {
  size_t N, N_post;
  int post_proc_flag;
} t_nb_iters;
static int cmp_nb_iters (const void *m1, const void *m2)
{
  t_nb_iters *n1 = (t_nb_iters *) m1;
  t_nb_iters *n2 = (t_nb_iters *) m2;
  return n1-> N > n2-> N;
}

size_t yaStoSOO_get_N_values (void)
{
#ifdef ENABLE_NLOPT
  static size_t N_values = 2 * 14;
#else
  static size_t N_values = 14;
#endif
  return N_values;
}
/* the next lines are a hack for CEC competition */
double les_coeff [] = { 0.01, 0.02, 0.03, 0.05, 0.1, 0.2, 0.3,
			0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0 };

void yaStoSOO_optim (void)
{
  size_t x_star_id = 0;
#ifdef ENABLE_NLOPT
  double *x_star = Malloc (sizeof (double) * d);
#endif
  t_nb_iters *nb_iters;
  if (verbose_mode) 
    fprintf (stderr, "soo flag = %d\n", soo_flag);
  nb_iters = (t_nb_iters *) Malloc (sizeof (t_nb_iters) * 
				    yaStoSOO_get_N_values ());
  les_x_star_val = 
    (t_les_x_star_val *) Malloc (sizeof (t_les_x_star_val) * 
				 yaStoSOO_get_N_values ());
  for (size_t i = 0; i < yaStoSOO_get_N_values (); i ++) {
    les_x_star_val [i]. x_star = Malloc (sizeof (double) * d);
    les_x_star_val [i]. x_star_preproc = Malloc (sizeof (double) * d);
  }
#ifdef ENABLE_NLOPT
  for (size_t i = 0; i < yaStoSOO_get_N_values () / 2; i ++) {
    nb_iters [2 * i]. N = les_coeff [i] * N;
    nb_iters [2 * i]. N_post = 0;
    nb_iters [2 * i]. post_proc_flag = 0;
    nb_iters [2 * i + 1]. N = les_coeff [i] * N * stosoo_percentage;
    nb_iters [2 * i + 1]. N_post = les_coeff [i] * N * (1. - stosoo_percentage);
    nb_iters [2 * i + 1]. post_proc_flag = 1;
  }
#else
  for (size_t i = 0; i < yaStoSOO_get_N_values (); i ++) {
    nb_iters [i]. N = les_coeff [i] * N;
    nb_iters [i]. N_post = 0;
    nb_iters [i]. post_proc_flag = 0;
  }
#endif
  qsort (nb_iters, yaStoSOO_get_N_values (), sizeof (t_nb_iters), cmp_nb_iters);
  for (size_t i = 0; i < yaStoSOO_get_N_values (); i ++) {
    while (nb_calls_to_obj_fct < nb_iters [i]. N) {
      int i_did_not_do_anything = 1;
      double bmax = -HUGE_VAL;
      for (size_t depth = hmin;
	   depth <= min (current_hmax, hmax - 2); depth ++) {
	/* Find the leaf to deal with at this depth */
	double local_bmax = -HUGE_VAL;
	size_t the_one = 0;
	for (size_t leaf_index = 0; leaf_index < nb_leaves [depth]; 
	     leaf_index ++) {
	  /* ties?: . */
	  size_t node = leaves [depth] [leaf_index];
	  if (is_a_leaf (node) && (nodes [node]. b > local_bmax)) {
	    the_one = node;
	    local_bmax = nodes [node]. b;
	  }
	}
	/* if the_one != 0, then the_one is the index of the node that:
	   - is a leaf
	   - is at current depth
	   - maximizes the confidence bound
	*/
	/*	if (the_one) {*/
	  if (local_bmax > bmax) {
	    if (nodes [the_one]. nb_eval < k) { /* re-evalute */
	      double f_of_x = f (nodes [the_one]. center);
	      nodes [the_one]. nb_eval ++;
	      nodes [the_one]. sum_eval += f_of_x;
	      nodes [the_one]. b = 
		nodes [the_one]. sum_eval / nodes [the_one]. nb_eval + 
		(soo_flag ? .0 : 
		 sqrt (log (N * k / delta) / 2.0 / nodes [the_one]. nb_eval));
	      i_did_not_do_anything = 0;
	    } else if ((nodes [the_one]. depth != hmax)
		       && (nb_calls_to_obj_fct + K - 1 < N)) /* assumes K odd */
	      { add_to_split_list (the_one); bmax = local_bmax; }
	  }
	  /*	} else fprintf (stderr, "set hmin <- %ld\n", hmin = depth + 1);*/
      } /* The whole tree has been processed at each depth. */
      if (nb_nodes_in_the_split_list) split_nodes ();
      else if (i_did_not_do_anything) {
	fprintf (stderr, "Nothing done. Breaking.\n");
	break;
      }
      
    }
    x_star_val = nodes [0]. sum_eval / nodes [0]. nb_eval;
    for (size_t node = 1; node < nb_nodes; node ++)
      if (nodes [node]. nb_eval && 
	  (nodes [node]. sum_eval / nodes [node]. nb_eval > x_star_val)) {
	x_star_val = nodes [node]. sum_eval / nodes [node]. nb_eval;
	x_star_id = node;
      }
    les_x_star_val [i]. x_star_val = 
      les_x_star_val [i]. x_star_val_preproc = x_star_val;
    les_x_star_val [i]. stosoo_nb_eval = nb_iters [i]. N;
    memcpy (les_x_star_val [i]. x_star_preproc, 
	    nodes [x_star_id]. center,
	    sizeof (double) * d);
#ifdef ENABLE_NLOPT
    if ((stosoo_percentage < 1.0) && (nb_iters [i]. post_proc_flag)) {
      (void) memcpy (x_star, nodes [x_star_id]. center, sizeof (double) * d);
      les_x_star_val [i]. post_proc_flag = 1;
      les_x_star_val [i]. post_proc_nb_eval = nb_iters [i]. N_post;
      les_x_star_val [i]. x_star_val = 
	post_process (x_star, les_x_star_val [i]. post_proc_nb_eval);
      memcpy (les_x_star_val [i]. x_star, x_star, sizeof (double) * d);
    } else {
      les_x_star_val [i]. post_proc_flag = 
	les_x_star_val [i]. post_proc_nb_eval = 0;
      memcpy (les_x_star_val [i]. x_star, nodes [x_star_id]. center,
	      sizeof (double) * d);
    }
#endif
  }
}

double *yaStoSOO_x_star (void)
{
  return x_star;
}
double yaStoSOO_x_star_val (void)
{ 
  return x_star_val;
}
